import 'dart:collection';

import 'package:yuro_core/yuro_core.dart';
import 'package:yuro_extension/yuro_extension.dart';

typedef InstanceBuilder<T> = T Function();

typedef InstanceAsyncBuilder<T> = Future<T> Function();

class YuroInstance {
  static YuroInstance _instance;

  factory YuroInstance() => _instance ??= YuroInstance._();

  const YuroInstance._();

  /// 保存使用[Yuro.put()]时创建的实例的引用
  static final Map<String, _InstanceBuilderFactory> _singleMap = {};

  /// 保存使用[Yuro.lazyPut()]时创建的实例的引用
  static final Map<String, _InstanceLazyBuilderFactory> _lazyMap = {};

  static final Map<String, String> _routeKeys = {};

  static final Map<String, HashSet<Function>> _routeByCreate = {};

  T put<T>(T dependency, {String tag, bool permanent = false}) {
    _insert(() => dependency, tag, true, permanent);
    return find<T>(tag: tag);
  }

  Future<T> putAsync<T>(InstanceAsyncBuilder<T> builder, {String tag, bool permanent = false}) async {
    return put<T>(await builder(), tag: tag, permanent: permanent);
  }

  void lazyPut<T>(InstanceBuilder<T> builder, {String tag, bool permanent = false}) {
    final key = _getKey(T, tag);
    _lazyMap.putIfAbsent(key, () => _InstanceLazyBuilderFactory(builder, permanent));
  }

  void create<T>(InstanceBuilder<T> builder, {String tag, bool permanent = true}) {
    _insert(builder, tag, false, permanent);
  }

  void _insert<T>(InstanceBuilder<T> builder, String tag, bool isSingle, bool permanent) {
    assert(builder != null);
    final key = _getKey(T, tag);
    _singleMap.putIfAbsent(key, () => _InstanceBuilderFactory<T>(builder, permanent, isSingle, false));
  }

  String _getKey(Type type, String tag) => tag == null ? type.toString() : '${type.toString()}_$tag';

  T find<T>({String tag}) {
    final key = _getKey(T, tag);
    if (isRegistered<T>(tag: tag)) {
      if (_singleMap[key] == null) {
        if (tag == null) {
          throw '类"$T"未注册.';
        } else {
          throw '带有标签"$tag"的类"$T"未注册.';
        }
      }
      final instance = _initDependency<T>(tag: tag);
      return instance ?? _singleMap[key].dependency as T;
    } else {
      if (!isPrepared<T>(tag: tag)) {
        throw '类"$T"未找到,需要调用"Yuro.put($T())"或者"Yuro.lazyPut(()=>$T())"注册.';
      }
      Yuro.log('懒加载实例"$T"已创建.');
      final value = put<T>(_lazyMap[key].builder() as T, tag: tag);
      _initDependency<T>(tag: tag);
      if (Yuro.smartManagement != SmartManagement.keepFactory && !_lazyMap[key].permanent) {
        _lazyMap.remove(key);
      }
      return value;
    }
  }

  bool isRegistered<T>({String tag}) => _singleMap.containsKey(_getKey(T, tag));

  bool isPrepared<T>({String tag}) => _lazyMap.containsKey(_getKey(T, tag));

  bool reset({bool clearFactory = true, bool clearRoute = true}) {
    if (clearFactory) _lazyMap.clear();
    if (clearRoute) _routeKeys.clear();
    _singleMap.clear();
    return true;
  }

  Future<bool> delete<T>({String tag, String key, bool force = false}) async {
    final newKey = key ?? _getKey(T, tag);
    if (!_singleMap.containsKey(newKey)) {
      Yuro.log('"$newKey"实例已被移除.');
      return false;
    }
    final instance = _singleMap[newKey];
    if (instance.permanent && !force) {
      Yuro.log('$newKey已被标记为持久保存,无权删除!');
      return false;
    }
    final dependency = instance.dependency;
    if (dependency is YuroServiceMixin && !force) {
      return false;
    }

    if (dependency is YuroLifeCycle) {
      dependency.onDelete();
      Yuro.log('$key的"onDispose()"被调用.');
    }
    _singleMap.removeWhere((k, v) => k == newKey);
    if (_singleMap.containsKey(newKey)) {
      Yuro.log('$newKey移除失败!');
    } else {
      Yuro.log('$newKey已从内存中移除.');
    }
    return true;
  }

  T _initDependency<T>({String tag}) {
    final key = _getKey(T, tag);
    final isInit = _singleMap[key].isInit;
    T t;
    if (!isInit) {
      t = _startController<T>(tag: tag);
      if (_singleMap[key].isSingle) {
        _singleMap[key].isInit = true;
        if (Yuro.smartManagement != SmartManagement.onlyBuilder) {
          _registerRoute(tag: tag);
        }
      }
    }
    return t;
  }

  T _startController<T>({String tag}) {
    final key = _getKey(T, tag);
    final instance = _singleMap[key].dependency as T;
    if (instance is YuroLifeCycleBase) {
      if (instance.onStart != null) {
        instance.onStart();
      }
      if (!_singleMap[key].isSingle && instance.onDelete != null) {
        _routeByCreate[Yuro.reference] ??= HashSet<Function>();
        _routeByCreate[Yuro.reference].add(instance.onDelete);
      }
    }
    return instance;
  }

  void _registerRoute<T>({String tag}) {
    _routeKeys.putIfAbsent(_getKey(T, tag), () => Yuro.reference);
  }

  void removeDependencyByRoute(String routeName) {
    final keysToRemove = _routeKeys.where((k, v) => v == routeName).keys.toList();

    if (_routeByCreate.containsKey(routeName)) {
      _routeByCreate[routeName].forEach((onDelete) => onDelete?.call());
      _routeByCreate[routeName].clear();
      _routeByCreate.remove(routeName);
    }

    keysToRemove.forEach((element) {
      delete(key: element);
      _routeKeys.remove(element);
    });

    keysToRemove.clear();
  }
}

class _InstanceBuilderFactory<T> {
  /// 标记是否为单例
  bool isSingle;

  /// 当[isSingle]=true时,存储该对象
  T _dependency;

  /// 当[isSingle]=false时,存储生成实例的工厂方法
  InstanceBuilder<T> builder;

  /// 是否持久化
  bool permanent;
  bool isInit;

  _InstanceBuilderFactory(
    this.builder,
    this.permanent,
    this.isSingle,
    this.isInit,
  );

  T get dependency => isSingle ? _dependency ??= builder() : builder();
}

class _InstanceLazyBuilderFactory<T> {
  final InstanceBuilder<T> builder;
  final bool permanent;

  _InstanceLazyBuilderFactory(this.builder, this.permanent);
}
